import re
import warnings
from pathlib import Path
from typing import Any

import numpy as np

INP_FILE_TEMPLATE = """\
cell  {cell}
alat  {alat}
ncell {ncell}
bc    P P P
sym   0

posfile     ./posfile
posfiletype {posfiletype}
momfile     ./momfile
exchange    ./exchange
maptype     {maptype}

initmag {initmag}
{restartfile_line}

ip_mode    M
ip_temp    {ip_temp}
ip_mcnstep {ip_mcnstep}

mode    M
temp    {temp}
mcnstep {mcnstep}

plotenergy   1
do_proj_avrg Y
do_cumu      Y
"""


def serialize_parameters(parameters: dict[str, Any]) -> dict[str, str]:
    """Serialize parameters for input file and validate selected.

    Most parameters are simply converted to string by calling `str(value)`. No further
    checks are performed for these.

    The following parameters are treated special:
    - cell: check that it is a 3x3 matrix of numbers and convert to a multi-line string
    - ncell: check that it is a length 3 array_like and convert to a strin in uppasd
          format (elements separated by spaces, no brackets)
    """
    serialized = {}
    for key, val in parameters.items():
        if key == "cell":
            val = _serialise_cell(val)
        elif key == "ncell":
            val = _serialize_array("ncell", val)
        serialized[key] = str(val)

    return serialized


def _serialise_cell(val: Any) -> str:
    cell = np.asanyarray(val)  # captures all shape mismatches
    if cell.dtype not in [int, float]:
        raise TypeError(
            f"'cell' elements must be of type int or float, not {cell.dtype}"
        )
    if cell.shape != (3, 3):
        raise ValueError(
            f"'cell' must be a 3x3 matrix or a list of 3 vectors; "
            f"got incompatible shape '{cell.shape}'"
        )
    # additional indentation for lines 2 and 3 to account for 'cell  ' in the
    # first line
    return (
        f"{cell[0, 0]} {cell[0, 1]} {cell[0, 2]}\n"
        f"      {cell[1, 0]} {cell[1, 1]} {cell[1, 2]}\n"
        f"      {cell[2, 0]} {cell[2, 1]} {cell[2, 2]}"
    )


def _serialize_array(name: str, val: Any, length: int = 3) -> str:
    if len(val) != length:
        raise ValueError(f"'{name}' must be of length {length}, not {len(val)}.")
    return " ".join(map(str, val))


def create_input_files(simulation_parameters) -> tuple[str, dict[str, Path]]:
    """Create inpsd.dat content and collect auxiliary files.

    Returns:
        - The content of inpsd.dat as a string.
        - A dictionary of auxilary files; keys are the names used in the run directory
            and the inpsd.dat file, values are `Path`s to the original files.
    """
    # first process all files; the `external_file` function modifies kwargs
    # to remove user-given paths to files as they have hard-coded names in the
    # inpsd.dat template

    # hard-coded names for exchange, posfile, momfile
    # these three are always required, other files are not supported in the Python
    # interface
    files_to_copy = {}
    for file_ in ["exchange", "posfile", "momfile"]:
        files_to_copy[file_] = _external_file(simulation_parameters, file_)

    if simulation_parameters.get("initmag") == 4:
        file_ = "restartfile"
        if file_ not in simulation_parameters:
            raise AttributeError("restartfile for initmag 4 missing")
        files_to_copy[file_] = _external_file(simulation_parameters, file_)
        # restartfile information needs to be passed as a whole line to the template
        simulation_parameters["restartfile_line"] = "restartfile ./restartfile"
    else:
        simulation_parameters["restartfile_line"] = ""

    inp_file = INP_FILE_TEMPLATE.format(**serialize_parameters(simulation_parameters))

    return inp_file, files_to_copy


def _external_file(simulation_parameters: dict[str, Any], key: str) -> Path:
    """Find file passed for UppASD argument key.

    The key is removed from simulation_parameters, because all files have hard-coded
    relative paths and names in the input file template.
    """
    try:
        file_path = simulation_parameters.pop(key)
    except KeyError:
        raise AttributeError(f"Missing argument '{key}'.") from None
    if not isinstance(file_path, str | Path):
        raise ValueError(
            f"Invalid type for '{key}': must be str or Path, not {type(file_path)}."
        )
    file_path = Path(file_path)
    if not file_path.is_file():
        raise ValueError(f"File '{file_path!s}' passed for '{key}' does not exist.")

    return file_path


def preprocess_inpsd(
    inpsd: Path, simulation_parameters: dict[str, Any]
) -> tuple[str, dict[str, Path]]:
    """Preprocess an existing inpsd.dat file.

    Users can overwrite lines in the input file by passing the respective key and a
    new value. Comments in the old lines will be lost.

    Names for posfile, momfile and exchange are always replaced with the defaults.
    If they are not in simulation_parameters, their values are read from the given
    inpsd file.

    Most arguments allowed in the inpsd.dat template are supported. In particular cell,
    ncell and bc are handled correctly, BUT they must be given in lower case.

    Currently unsupported are multi-line options such as ip_mcanneal. Passing these as
    in simulation_parameters will result in syntactically wrong input files without
    any notice. Running a simulation with such a file should typically cause some sort
    of failure of UppASD, so the user should notice.

    Returns:
        - The content of inpsd.dat as a string.
        - A dictionary of auxilary files; keys are the names used in the run directory
            and the inpsd.dat file, values are `Path`s to the original files.
    """
    inp_file = inpsd.read_text()
    files_to_copy = {}
    for key, val in simulation_parameters.items():
        pattern = rf"^{key}\s.*$"
        if key == "cell":
            val = _serialise_cell(val) + "\n"  # TODO is this newline required?
            # replace three consecutive lines, where the first line starts with 'cell '
            pattern = rf"^{key}\s.*\n.*\n.*$"  # TODO write a test for this
        elif key in ["ncell", "bc"]:
            val = _serialize_array(key, val)

        val = str(val)

        if "\\" in val:
            # paths on Windows contain backslashes
            val = re.escape(val)
        inp_file = re.sub(pattern, rf"{key} {val}", inp_file, flags=re.MULTILINE)

    # only the following four files are supported, restartfile is only dealt with
    # properly when initmag is set to 4
    for file_ in ["exchange", "posfile", "momfile", "restartfile"]:
        if file_ not in simulation_parameters and (
            # find file name and path in input file, ignore trailing comment (not sure
            # if UppASD allows one in file lines)
            match_ := re.search(rf"^{file_}\s+([^\s+]+)", inp_file, flags=re.MULTILINE)
        ):
            # add file name from input file to simulation parameters to be able to use
            # the 'external_file' function
            simulation_parameters[file_] = match_.group(1)

        if file_ == "restartfile":
            init_mag = re.search(r"initmag\s+([^\s]+)", inp_file.lower())
            if not init_mag:
                raise RuntimeError("Missing option 'initmag' in inpsd.dat.")
            if int(init_mag.group(1)) != 4:
                # restartfile is only used for initmag 4, so we ignore it for any other
                # value of initmag
                if file_ in simulation_parameters or (match_ and match_.group(1)):
                    warnings.warn(
                        "The parameter 'restartfile' is ignored for initmag!=4, got"
                        f" initmag {init_mag.group(1)}.",
                        stacklevel=3,
                    )
                continue

        files_to_copy[file_] = _external_file(simulation_parameters, file_)
        # insert hard-coded names for auxilary files to ensure that each run directory
        # is fully self-contained
        inp_file = re.sub(
            rf"^{file_}\s.*$",
            f"{file_} ./{file_!s}",
            inp_file,
            flags=re.MULTILINE,
        )

    return inp_file, files_to_copy



import collections
import copy
import datetime
import numbers
import re
import shutil
import string
import subprocess
import warnings
from pathlib import Path
from typing import Any

import numpy as np
import yaml

import mammos_spindynamics

from . import _data
from ._inpsd import INP_FILE_TEMPLATE, create_input_files, preprocess_inpsd

DEFAULT_SIMULATION_PARAMETERS = {
    "ncell": [25, 25, 25],
    "ip_mcnstep": 25_000,
    "mcnstep": 50_000,
}


class Simulation:
    """Class to perform UppASD simulations.

    Simulations can be performed without or with existing inpsd.dat file:

    - If no inpsd.dat file is passed the simulation object will use an internal
      template to create a new inpsd.dat file. This template requires a number of
      parameters, which have to be passed by the user either when creating the
      `simulation` object, or when calling :py:func:`~Simulation.run` or
      :py:func:`~Simulation.temperature_sweep`. A list of  required parameters can be
      obtained from :py:func:`~Simulation.required_parameters`.
    - If an inpsd.dat file is passed most of its lines can be modified by passing a new
      value for a parameters as keyword argument. It is not possible to overwrite
      ``ip_mcanneal``.
    """

    def __init__(self, inpsd: Path | str | None = None, **kwargs):
        """Create a new simulation object with an inpsd.dat file and/or parameters.

        Args:
            inpsd: Optional path to an existing inpsd.dat file. If not passed a
                default template defined in the `Simulation` class is used.
            kwargs: Parameters to use (or overwrite when passing `inpsd`) in
                inpsd.dat. Parameters can also be passed when calling
                :py:func:`Simulation.run` or :py:func:`Simulation.temperature_sweep`.
        """
        self._inpsd = inpsd
        self._simulation_parameters_init = kwargs

    @property
    @staticmethod
    def allowed_parameters(self):
        """UppASD inpsd parameters that can be passed without a custom inpsd file."""
        return set(
            key
            for _text, key, _format_spec, _conversion in string.Formatter().parse(
                INP_FILE_TEMPLATE
            )
            if key is not None and key not in ["restartfile_line"]
        )

    @property
    @staticmethod
    def required_parameters(self):
        """UppASD inpsd parameters that must be passed without a custom inpsd file."""
        return set(
            key
            for key in self.allowed_parameters
            if key not in DEFAULT_SIMULATION_PARAMETERS
        )

    def __repr__(self):
        args = "".join(
            f"    {key}={val!r},\n" for key, val in self._defined_parameters().items()
        )
        inpsd = f"    inpsd={self._inpsd},\n" if self._inpsd else ""
        return f"{self.__class__.__name__}(\n{inpsd}{args})"

    def _defined_parameters(self, **kwargs) -> dict[str, Any]:
        """UppASD parameters that are currently defined.

        Merges parameters passed at object creation time, parameters passed as kwargs
        to this method and optionally default parameters if no custom inpsd.dat file
        is used.

        A copy of the dictionary is returned, to ensure that neither default parameters
        nor parameters passed at object creation time are modified.
        """
        if self._inpsd:
            simulation_parameters = copy.copy(self._simulation_parameters_init)
            simulation_parameters.update(kwargs)
        else:
            simulation_parameters = copy.copy(DEFAULT_SIMULATION_PARAMETERS)
            simulation_parameters.update(copy.copy(self._simulation_parameters_init))
            simulation_parameters.update(kwargs)
        return simulation_parameters

    def create_input_files(self, **kwargs) -> tuple[str, dict[str, Path]]:
        """Create input files required for UppASD: inpsd.dat and auxilary files.

        This method creates the content for inpsd.dat and collects auxiliary files
        `posfile`, `momfile`, `exchange`, and optionally `restartfile`. Other external
        files are not supported.

        If no custom inpsd.dat file is used, a predefined template will be used.
        The user has to pass all required template arguments either at object creation
        time or when calling this method. A list of required arguments can be obtained
        from :py:func:`Simulation.required_parameters`.

        If a custom inpsd.dat file is used, the user has the option to overwrite lines
        in that input files by passing the respective keys and new values. Not all
        options in inpsd.dat are currently supported, in particular `ip_mcanneal` cannot
        be used.

        The auxiliary files have hard-coded names in inpsd.dat, when passing a custom
        inpsd.dat the corresponding lines will be modified (trailing comments are lost).
        The dict returned as second object maps names used in the inpsd.dat file to
        original file paths.
        """
        simulation_parameters = self._defined_parameters(**kwargs)

        if T := simulation_parameters.pop("T", None):
            # convenience for the user: set both ip_temp and temp to the same value
            if "ip_temp" in simulation_parameters or "temp" in simulation_parameters:
                raise ValueError(
                    "Parameter 'T' cannot be used simultaneously with parameters"
                    " '(temp, ip_temp)'"
                )
            simulation_parameters["ip_temp"] = T
            simulation_parameters["temp"] = T

        if self._inpsd:
            return preprocess_inpsd(Path(self._inpsd), simulation_parameters)
        else:
            if missing_parameters := self.required_parameters - set(
                simulation_parameters.keys()
            ):
                raise RuntimeError(
                    f"The following parameters are missing: {missing_parameters}"
                )
            return create_input_files(simulation_parameters)

    def run(
        self,
        out: str | Path,
        description: str = "",
        uppasd_executable: str | Path = "uppasd",
        verbosity: int = 1,
        **kwargs,
    ) -> _data.RunData:
        """Run a single UppASD simulation.

        This method creates inputs required for UppASD in a new directory, runs UppASD
        in that directory and returns an object that allows accessing the resulting
        data.

        Args:
            out: Base directory in which the output is stored. The method will
                create a subdirectory ``<index>-run`` inside that directory. The value
                of ``<index>`` depends on the existing directory content. For empty
                directories ``<index>=0``. Otherwise, the method will search for all
                existing ``<some-index>-run`` and ``<some-index>-temperature_sweep``
                directories and use the next index.
            description: Human-readable description of this simulation run, stored as
                metadata.
            uppasd_executable: Name or path to UppASD executable. If provided a name
                the method will search for the executable on PATH.
            verbosity: Verbosity of the run:

                - 0: no output
                - 1: summary line with status information (output directory, runtime)
            kwargs: UppASD arguments to use in the inpsd.dat file, including paths to
                files for momfile, posfile, and exchange. Further details are explained
                in :py:func:`Simulation.create_input_files`.

                In addition a special argument ``T`` is supported, which will be used to
                set both ``ip_temp`` and ``temp``. Passing both ``T`` and (at least) one
                of ``ip_temp`` or ``temp`` is not supported.

                Parameters passed to `run` instead of at object creation time are
                recorded as separate metadata and can be used later on to conveniently
                retrieve simulation results.

        Returns:
            An object to access UppASD outputs.
        """
        out = Path(out)
        uppasd_executable = _find_executable(uppasd_executable)

        inp_file_content, files_to_copy = self.create_input_files(**kwargs)

        run_path, index = _create_run_dir(out)

        metadata = {
            "metadata": {
                "mammos_spindynamics_version": mammos_spindynamics.__version__,
                "mode": "run",
                "description": description,
                "index": index,
            },
            "parameters": {key: _sanitize_val(value) for key, value in kwargs.items()},
        }
        _write_inputs(run_path, inp_file_content, files_to_copy, metadata)

        time_start = datetime.datetime.now().isoformat(timespec="seconds")
        if verbosity == 1:
            print(f"Running UppASD in {run_path!s} ...", end=" ")
        _run_simulation(run_path, uppasd_executable)
        time_end = datetime.datetime.now().isoformat(timespec="seconds")
        time_elapsed = datetime.datetime.fromisoformat(
            time_end
        ) - datetime.datetime.fromisoformat(time_start)

        if verbosity == 1:
            print(f"simulation finished, took {time_elapsed!s}")

        _update_metadata_file(run_path, time_start, time_end, time_elapsed)
        return _data.RunData(run_path)

    def temperature_sweep(
        self,
        T: collections.abc.Iterable[numbers.Real],
        out: str | Path,
        restart_with_previous: bool = True,
        description: str = "",
        uppasd_executable: str | Path = "uppasd",
        verbosity: int = 2,
        **kwargs,
    ) -> _data.TemperatureSweepData:
        """Run temperature sweep.

        This method runs consecutive simulations for multiple temperatures. For each
        temperature it will call :py:func:`Simulation.run` internally. The method does
        not sort the temperatures, so make sure you pass the desired order, in
        particular when using `restart_with_previous=True`.

        After finishing all simulations two additional files `M(T)` and `output.csv`
        with aggregated simulation results are created.

        Args:
            T: Temperatures in Kelvin to run the simulation for. The value will be used
                for both ``ip_temp`` and ``temp``.
            out: Base directory for the output. A new directory
                ``<index>-temperature_sweep`` will be created inside, the logic is
                equivalent to :py:func:`Simulation.run`.
            restart_with_previous: If set to True use the restart file from the previous
                run as initial configuration for all but the first run in the sweep.
            description: Human-readable description of the sweep, the individual runs
                do not have a description.
            uppasd_executable: Name or path to UppASD executable, like in
                :py:func:`Simulation.run`.
            verbosity: Verbosity of the sweep:

                - 0: no output
                - 1: only summary output for the whole sweep (number and list of
                  temperatures)
                - 2: summary for the sweep and summary for each run (verbosity 1 for
                  each run)
            kwargs: UppASD arguments in inpsd.dat as in :py:func:`Simulation.run`. The
                two options ``ip_temp`` and ``temp`` cannot be used as instead ``T`` is
                required.

        Returns:
            An object to access UppASD outputs.

        """
        run_path, index = _create_run_dir(Path(out), mode="temperature_sweep")

        # convert any form of T to a list of T values
        Ts = np.asanyarray(T).tolist()

        metadata = {
            "metadata": {
                "description": description,
                "index": index,
                "mode": "temperature_sweep",
            },
            "parameters": {
                "T": Ts,
                **{key: _sanitize_val(value) for key, value in kwargs.items()},
            },
        }
        with open(run_path / "mammos_spindynamics.yaml", "w") as f:
            yaml.dump(metadata, f)

        if verbosity >= 1:
            print(
                f"Running simulations for {len(Ts)} different temperatures:\n    {Ts!s}"
            )

        # run first simulation with default options; later simulations optionally with
        # restarting from previous
        if verbosity >= 2:
            print(f"T={Ts[0]}:", end=" ")
        run_data = self.run(
            T=Ts[0],
            out=run_path,
            uppasd_executable=uppasd_executable,
            verbosity=verbosity - 1,
            **kwargs,
        )
        for T_ in Ts[1:]:
            if restart_with_previous:
                kwargs.update({"initmag": 4, "restartfile": run_data.restartfile})
            if verbosity >= 2:
                print(f"T={T_}:", end=" ")
            run_data = self.run(
                T=T_,
                out=run_path,
                uppasd_executable=uppasd_executable,
                verbosity=verbosity - 1,
                **kwargs,
            )

        result = _data.TemperatureSweepData(run_path)
        result.save_output(run_path)
        return result


def _create_run_dir(base: Path, mode="run") -> tuple[Path, int]:
    """Create a new directory for run or temperature_sweep.

    The directory name is prefixed with an index, which enumerates all runs and
    temperature_sweeps. The function looks for the maximum existing index and uses
    the next index.
    """
    if mode not in ["run", "temperature_sweep"]:
        raise ValueError(
            f"Mode {mode} not supported, must be 'run' or 'temperature_sweep'"
        )

    if not base.exists():
        base.mkdir(parents=True)
    elif base.is_file():
        raise RuntimeError(f"The path '{base}' passed as output directory is a file.")

    if not (base / "mammos_spindynamics.yaml").exists():
        with open(base / "mammos_spindynamics.yaml", "w") as f:
            yaml.dump({"metadata": {"mode": "mammos_uppasd_data"}}, f)

    run_indices = [
        int(p.name.split("-")[0])
        for p in base.iterdir()
        if re.match(r"^\d+-(run|temperature_sweep)$", p.name)
    ]
    next_index = max(run_indices) + 1 if run_indices else 0

    next_run_path = base / f"{next_index}-{mode}"
    # The next call would fail if the directory exists already. This should never
    # happen. We can rely on it as additional safety-check to not overwrite anything.
    next_run_path.mkdir()
    return next_run_path, next_index


def _write_inputs(
    out_path: Path,
    inp_file_content: str,
    files_to_copy: dict[str, Path],
    metadata: dict[str, Any],
) -> None:
    """Write input files.

    The function writes the file inpsd.dat, copies all files in ``files_to_copy``,
    where keys are the new filenames and values are paths to the original files, and
    create a ``mammos_spindynamics.yaml`` file with ``metadata``.

    The directory ``out_path`` must already exist.
    """
    (out_path / "inpsd.dat").write_text(inp_file_content)
    for name, orig_path in files_to_copy.items():
        shutil.copy(orig_path, out_path / name)
    with open(out_path / "mammos_spindynamics.yaml", "w") as f:
        yaml.dump(metadata, f)


def _find_executable(uppasd_executable: str) -> Path:
    """Find executable with given name in PATH."""
    exe = shutil.which(uppasd_executable)
    if not exe:
        raise RuntimeError(
            f"Could not find UppASD executable with name '{uppasd_executable}' in PATH"
        )
    return Path(exe).resolve()


def _run_simulation(run_dir: Path, uppasd_executable: str):
    """Call UppASD in ``run_dir``.

    Stdout and stderr are redirected to files ``uppasd_stdout.txt`` and
    ``uppasd_stderr.txt``. The method will raise an exception if UppASD fails and print
    a warning if stdout contains the word ERROR, which is often the case when running
    a simulation with an invalid input file (UppASD does typically not fail in such
    cases).
    """
    res = subprocess.run(
        uppasd_executable,
        cwd=run_dir,
        check=True,
        capture_output=True,
        text=True,
    )

    if res.stderr:
        warnings.warn(
            f"Stderr is not empty, the simulation may have failed:\n{res.stderr}",
            stacklevel=3,
        )

    if "ERROR" in res.stdout:
        warnings.warn(
            "UppASD output contains ERROR lines, simulation likely failed:\n"
            + res.stdout,
            stacklevel=3,
        )

    (run_dir / "uppasd_stdout.txt").write_text(res.stdout)
    (run_dir / "uppasd_stderr.txt").write_text(res.stderr)


def _update_metadata_file(
    run_path: Path, time_start: str, time_end: str, time_elapsed: str
) -> None:
    """Update ``mammos_spindynamics.yaml`` metadata file.

    Inserts:
      - start/end/elapsed time
      - if available the git revision reported by UppASD
    """
    # convert to string first to limit resolution to seconds
    with open(run_path / "mammos_spindynamics.yaml") as f:
        metadata = yaml.safe_load(f)

    uppasd_yaml = list(run_path.glob("uppasd.*.yaml"))
    uppasd_yaml = None
    if uppasd_yaml:
        with open(uppasd_yaml[0]) as f:
            uppasd_git_revision = yaml.safe_load(f)["git_revision"]
    else:
        uppasd_git_revision = "<unknown>"

    metadata["metadata"].update(
        {
            "time_start": time_start,
            "time_end": time_end,
            "time_elapsed": str(time_elapsed),
            "uppasd_git_revision": uppasd_git_revision,
        }
    )
    with open(run_path / "mammos_spindynamics.yaml", "w") as f:
        yaml.dump(metadata, f)


def _sanitize_val(val: Any) -> Any:
    """Convert some types to improve yaml dump.

    Converts:
    - tuple -> list
    - Path -> str
    """
    # TODO replace with custom YamlDumper
    if isinstance(val, Path):
        return str(val)
    if isinstance(val, tuple):
        return list(val)
    return val



"""Test db lookup."""

import pathlib

import numpy as np
import pytest

from mammos_spindynamics.db import get_spontaneous_magnetization

DATA_DIR = pathlib.Path(__file__).parent.resolve() / "data"


def test_Co2Fe2H4():
    """Test material `Co2Fe2H4`.

    There is only one material with formula `Co2Fe2H4`, so this
    test should load its table without issues.
    """
    magnetization_data = get_spontaneous_magnetization(chemical_formula="Co2Fe2H4")
    assert np.allclose(magnetization_data.T.value, magnetization_data.dataframe["T"])
    assert np.allclose(magnetization_data.Ms.value, magnetization_data.dataframe["Ms"])


def test_NdFe14B():
    """Test material `NdFe14B`.

    There is no material with such formula in the database,
    so we expect a `LookupError`.
    """
    with pytest.raises(LookupError):
        get_spontaneous_magnetization(chemical_formula="NdFe14B")


def test_Co2Fe2H4_11():
    """Test material `Co2Fe2H4` with space group number 11.

    There is no material with such formula and space group
    in the database, so we expect a `LookupError`.
    """
    with pytest.raises(LookupError):
        get_spontaneous_magnetization(
            chemical_formula="Co2Fe2H4", space_group_number=11
        )


def test_all():
    """Test search with no filters.

    This will select all entries in the database,
    so we expect a `LookupError`.
    """
    with pytest.raises(LookupError):
        get_spontaneous_magnetization()



"""Spindynamics-based magnetic material properties."""

import importlib.metadata

from mammos_spindynamics import db, uppasd

__all__ = ["db", "uppasd"]

__version__ = importlib.metadata.version(__package__)
